#include <iostream>
#include "Polymers.h"
#include "defs.h"

//#define AMOUNT_LEFT

Polymer::Polymer(int NumMon, double StepLength, double Repulsion, long seed){
   set(NumMon, StepLength, Repulsion);
   //Rand = new RandNumGenerator(seed);
   min_head = NumMon-1;
   max_tail = 0;
}

Polymer::Polymer(){
   Polymer(0,0.0,0.0,0L);
}

void Polymer::putBinders(MyosinBinding * Bind1, MyosinBinding * Bind2){
   this->Bind1 = Bind1;
   this->Bind2 = Bind2;
}

void Polymer::isMyosinBound(bool * b){
   b[0] = Bind1->isBound();
   b[1] = Bind2->isBound();
}

MyosinBinding * Polymer::getBind1(){
   return Bind1;
}

MyosinBinding * Polymer::getBind2(){
   return Bind2;
}


void Polymer::setSpringConstant(double SpringConstant){
   this->SpringConstant = SpringConstant;
}

void Polymer::setSpringConstantArray(double * SpringConstantArray){
   this->SpringConstantArray = SpringConstantArray;
}

void Polymer::setNNConstant(double NNConstant){
   this->NNConstant = NNConstant;
}


void Polymer::setExtForce(double ExtForce){
   this->ExtForce = ExtForce;
}

void Polymer::setCoords(vect_f * Coords){
   this->Coords = Coords;
}

void Polymer::setVelocities(vect_f * Velocities){
   this->Velocities = Velocities;
}


void Polymer::setParams(int NumMon, double StepLength, double Repulsion){
   MoveSize = 1;
   Beta = 1;
   clamped=false;
   ClampStrength=1;
   this->NumMon = NumMon;
   this->StepLength = StepLength;
   this->seed = seed;
   this->Repulsion = Repulsion;
   Head=NumMon-1;
   min_head = NumMon-1;
   max_tail = 0;
   mark = new int[NumMon];
}

void Polymer::set(int NumMon, double StepLength, double Repulsion, vect_f * Coords){
   setParams(NumMon, StepLength, Repulsion);
   this->Coords = Coords;
}

void Polymer::set(int NumMon, double StepLength, double Repulsion){
   setParams(NumMon, StepLength, Repulsion);
   Coords = new vect_f[NumMon];
}

void Polymer::setPolymers(Polymers * polymers) {
   this->polymers = polymers;
   grid = polymers->grid;
   Rand = polymers->getRandGen();
}

void Polymer::setCores(double SoftCoreRadius, double HardCoreRadius) {
   SoftCoreRadiusSq = SoftCoreRadius*SoftCoreRadius;
   HardCoreRadiusSq = HardCoreRadius*HardCoreRadius;
   this->SoftCoreRadius = SoftCoreRadius;
   this->HardCoreRadius = HardCoreRadius;
   OneOverBlah = 1/(SoftCoreRadiusSq-HardCoreRadiusSq);
}

void Polymer::setActinParams(double VActinDepth, vect_f & act_dir, double ActinSpacing){
   actin_dir = act_dir;
   actin_dir /= sqrt(actin_dir*actin_dir);
   this->ActinSpacing = ActinSpacing;
   V_actin = VActinDepth/(0.125*QUART(SoftCoreRadiusSq));
}

vect_f Polymer::getActinDir() {
   return actin_dir;
}

void Polymer::setHeight(double Height){
   this->Height = Height;
}

double Polymer::getHeight(){
   return Height;
}


void Polymer::setActinSpacing(double ActinSpacing){
   this->ActinSpacing = ActinSpacing;
}

double Polymer::getActinSpacing() {
   return ActinSpacing;
}

void Polymer::setClamp(bool clamped) {
   this->clamped = clamped;
}

void Polymer::setClampStrength(double ClampStrength) {
   this->ClampStrength = ClampStrength;
   clamped = (ClampStrength > 0);
}

void Polymer::setClampedPositions(vect_f * ClampedPositions) {
   this->ClampedPositions = ClampedPositions;
}

void Polymer::setBeta(double Beta) {
   this->Beta = Beta;
}

void Polymer::setMoveSize(double MoveSize) {
   this->MoveSize = MoveSize;
}

int Polymer::getNumMon() {
   return NumMon;
}

vect_f * Polymer::array_copy(vect_f * C, int num) {
   vect_f * cpy = new vect_f[num];
   memcpy(cpy, C, num*sizeof(vect_f));
   return cpy;
}


vect_f * Polymer::getCoords() {
   return Coords;
}

void  Polymer::getXCoords(double * X) {
   for(int i=0; i < NumMon; i++) {
      X[i] = Coords[i].xcomp();
   }
}

void Polymer::getYCoords(double * Y) {
   for(int i=0; i < NumMon; i++) {
      Y[i] = Coords[i].ycomp();
   }
}

void Polymer::getShiftedCoords(double * X, double * Y) {
   vect_f * COM =  CenterOfMass();
   for(int i=0; i < NumMon; i++) {
      X[i] = Coords[i].xcomp() - COM->xcomp();
      Y[i] = Coords[i].ycomp() - COM->ycomp();
   }
   delete COM;
}
#ifdef GAUSSIAN
void Polymer::initRand(vect_f init_pos, vect_f BoxSize) {
   Coords[0] = init_pos;
   insert(0);
   double x=init_pos.xcomp();
   double y=init_pos.ycomp();
   double z=init_pos.zcomp();
   for(int i=1;i < NumMon;) {
      double tmp;
      tmp = x + StepLength * Rand->nextGaussian();
      if (tmp < 0 || tmp > BoxSize.xcomp())
	 continue;
      x = tmp;
      tmp = y + StepLength * Rand->nextGaussian();
      if (tmp < 0 || tmp > BoxSize.ycomp())
	 continue;
      y = tmp;
      tmp = z + StepLength * Rand->nextGaussian();
      if (tmp < 0 || tmp > BoxSize.zcomp())
	 continue;
      z = tmp;
      Coords[i].set(x,y,z);
      insert(i);
      i++;
   }
}

#else

void Polymer::initRand(vect_f init_pos, vect_f BoxSize) {
   Coords[0] = init_pos;
   insert(0);
   vect_f Rand3d;

   for(int i=1;i < NumMon;) {
      vect_f tmp;
      Rand->nextSphere(Rand3d);
      tmp = init_pos + StepLength * Rand3d;
//#define HARDWALLS
#ifdef HARDWALLS
      if (tmp.xcomp() < 0 || tmp.xcomp() > BoxSize.xcomp()
	    ||
         tmp.ycomp() < 0 || tmp.ycomp() > BoxSize.ycomp()
	    ||
         tmp.zcomp() < 0 || tmp.zcomp() > BoxSize.zcomp())
	 continue;
#endif
      Coords[i] = tmp;
      init_pos = tmp;
      insert(i);
      i++;
   }
}
#endif

void Polymer::initPutChain(double ** r) {
    for(int i=0;i < NumMon; i++) {
        Coords[i].set( r[i][0], r[i][1], r[i][2]);
        insert(i);
    }

}

void Polymer::initRand(vect_f init_pos, vect_f BoxSize, vect_f init_dir) {
   Coords[0] = init_pos;
   insert(0);
   vect_f Rand3d;

   for(int i=1;i < NumMon;) {
      vect_f tmp;
      Rand->nextSphere(Rand3d);
      Rand3d = -1.0*actin_dir + 0.1*Rand3d;
      double norm = sqrt(Rand3d*Rand3d);
      Rand3d /= norm;
      tmp = init_pos + StepLength * Rand3d;
      //#define HARDWALLS
#ifdef HARDWALLS
      if (tmp.xcomp() < 0 || tmp.xcomp() > BoxSize.xcomp()
	    ||
	    tmp.ycomp() < 0 || tmp.ycomp() > BoxSize.ycomp()
	    ||
	    tmp.zcomp() < 0 || tmp.zcomp() > BoxSize.zcomp())
	 continue;
#endif
      Coords[i] = tmp;
      init_pos = tmp;
      insert(i);
      i++;
   }
}

#if 0
void Polymer::initRand(vect_f init_pos, vect_f BoxSize, vect_f init_dir) {
   Coords[0] = init_pos;
   insert(0);
   vect_f Rand3d;

   for(int i=1;i < NumMon;) {
      vect_f tmp;
      if (i ==1 ) {
	 Rand3d = 1.0*actin_dir;
      }
      else {
	 Rand->nextSphere(Rand3d);
      }
      tmp = init_pos + StepLength * Rand3d;
//#define HARDWALLS
#ifdef HARDWALLS
      if (tmp.xcomp() < 0 || tmp.xcomp() > BoxSize.xcomp()
	    ||
         tmp.ycomp() < 0 || tmp.ycomp() > BoxSize.ycomp()
	    ||
         tmp.zcomp() < 0 || tmp.zcomp() > BoxSize.zcomp())
	 continue;
#endif
      Coords[i] = tmp;
      init_pos = tmp;
      insert(i);
      i++;
   }
}
#endif

void Polymer::initRand() {
   vect_f init_pos;
   vect_f boxwidth = grid->getBoxWidth();
   init_pos.set(boxwidth.xcomp()*Rand->nextFloat(),boxwidth.ycomp()*Rand->nextFloat(),boxwidth.zcomp()*Rand->nextFloat());
   initRand(init_pos, boxwidth);
}

void Polymer::initStraight(vect_f init_pos) {
   Coords[0] = init_pos;
   double x = init_pos.xcomp();
   double y = init_pos.ycomp();
   double z = init_pos.zcomp();
   for(int i=1;i < NumMon; i++) {
      x += StepLength;
      Coords[i].set(x,y,z);
      insert(i);
   }
}

void Polymer::initScrunched(vect_f init_pos) {
   Coords[0] = init_pos;
   double eps = 0.01;
   double x = init_pos.xcomp();
   double y = init_pos.ycomp();
   double z = init_pos.zcomp();
   double Dx = sqrt(SQR(StepLength) - SQR(eps));

   for(int i=1;i < NumMon; i++) {
      x += i%2 == 0 ? Dx: -Dx;
      y += eps;
      Coords[i].set(x,y,z);
      insert(i);
   }
}

void Polymer::initStraight() {
   vect_f init_pos;
   vect_f boxwidth = grid->getBoxWidth();
   init_pos.set(boxwidth.xcomp()*Rand->nextFloat(),boxwidth.ycomp()*Rand->nextFloat(),boxwidth.zcomp()*Rand->nextFloat());
   initStraight(init_pos);
}

void Polymer::initScrunched() {
   vect_f init_pos;
   vect_f boxwidth = grid->getBoxWidth();
   init_pos.set(boxwidth.xcomp()*Rand->nextFloat(),boxwidth.ycomp()*Rand->nextFloat(),boxwidth.zcomp()*Rand->nextFloat());
   initScrunched(init_pos);
}

void Polymer::init(vect_f * InitPositions) {
   for(int i=0;i < NumMon; i++) {
      Coords[i] = InitPositions[i];
   }
}

inline double Polymer::distsq(int i, int j) {
   return Coords[i].distsq(Coords[j]);
}

double Polymer::VClamp(vect_f & R,int Monomer) {
   return ClampStrength*ClampedPositions[Monomer].distsq(R);
}


inline double Polymer::EPointAndChainHard(vect_f & point, int monomer) {
   double e=0;
   for(int i=0;i < monomer;i++) {
      double dsq = point.distsq(Coords[i]);
      if (dsq < 1) e+= 100;
   }
   for(int i=monomer+1;i < NumMon;i++) {
      double dsq = point.distsq(Coords[i]);
      if (dsq < 1) e+= 100;
   }
   return e*Repulsion;
}

//single chain long range code:
#if 0
inline double Polymer::ePointAndChain(const vect_f & point, int monomer) const {
   double e=0;
   for(int i=0;i < monomer;i++) {
      //double dsq = point.distsq(Coords[i]);
      double dsq = (Coords[i]).distsq(point);
      e+= 1.0/(dsq*dsq);
   }
   for(int i=monomer+1;i < NumMon;i++) {
      //double dsq = point.distsq(Coords[i]);
      double dsq = (Coords[i]).distsq(point);
      e+= 1.0/(dsq*dsq);
   }
   return e*Repulsion;
}
#endif

double Polymer::ESpring() {
   double e=0;

   for(int i=0;i < NumMon-1; i++)
      e += Coords[i].distsq(Coords[i+1]);

   return e/2;
}

vect_f * Polymer::CenterOfMass() {
   vect_f * RcmP = new vect_f();
   for(int i=0; i < NumMon; i++) {
      RcmP[0] += Coords[i];
   }
   RcmP[0] /= (double) NumMon;
   return RcmP;
}

double Polymer::RgSq() {
   vect_f * RcmP = CenterOfMass();
   double rgsq=0;
   for(int i=0; i < NumMon; i++) {
      vect_f diff = Coords[i] - *RcmP;
      rgsq += diff*diff;
   }
   delete RcmP;
   return rgsq/NumMon;
}

double Polymer::REndEndSq() {
   return SQR(Coords[NumMon-1]-Coords[0]);
}

double Polymer::ERepulsive() {
   double e=0;

   for(int i=0;i < NumMon;i++) {
      for(int j=0;j < i;j++) {
	 double dsq = Coords[i].distsq(Coords[j]);
	 e+= 1.0/(dsq*dsq);
      }
   }
   return e;
}

int Polymer::fixedMove(int num_steps) {
   return reptate(num_steps);

   //return jiggle(num_steps);
}

void Polymer::rotate3Mons(int middle, vect_f & new_pos) {
   vect_f diff = 0.5*(Coords[middle+1]-Coords[middle-1]);
   double size = sqrt(diff*diff);
#if 0//check only necessary with no excluded volume
   if (size < 1e-6) {
      new_pos = Coords[middle];
      return;
   }
#endif
   
   vect_f norm_diff = diff/size;
   vect_f center = Coords[middle-1]+diff;
   vect_f perp = Coords[middle]-center;
   vect_f perp_perp = perp ^ norm_diff;
   double g1,g2;
   Rand->nextCircle(g1,g2);
   vect_f new_dir = g1 * perp + g2 * perp_perp;
   new_pos = center + new_dir;
}

int Polymer::jiggle_once() {
   vect_f  RNew;
   vect_f Rand3d;
   int NumAccepted=0;
   double *v = polymers->VConfining;
   double WellWidth = polymers->WellWidth;
#ifdef GAUSSIAN
   double Coeff = 0.5/(StepLength*StepLength);
#endif

   int Monomer = Rand->nextInt(NumMon);// Which monomer to move
   RNew = Coords[Monomer];

#ifdef GAUSSIAN
   Rand3d.set(MoveSize*Rand->nextGaussian(),MoveSize*Rand->nextGaussian(),MoveSize*Rand->nextGaussian());
   RNew += Rand3d;
#else
   if (Monomer == 0)
   {
      Rand->nextSphere(Rand3d);
      RNew = Coords[1] + StepLength*Rand3d;
   }
   else if (Monomer == NumMon-1)
   {
      Rand->nextSphere(Rand3d);
      RNew = Coords[NumMon-2] + StepLength*Rand3d;
   }
   else
      rotate3Mons(Monomer,RNew);
#endif

   if (!grid->inBox(RNew))
      return 0;

   remove(Monomer);
   //calculate Beta*Energy because spring term should be temperature
   //independent in Gibbs expononent
#ifdef GAUSSIAN
   double OldEnergy = ePointAndChains(Coords[Monomer],Monomer) + Coeff*bESpringNeighbors(Coords[Monomer],Monomer);
   double NewEnergy = ePointAndChains(RNew,Monomer) + Coeff*bESpringNeighbors(RNew,Monomer);
#else//fixed length
   double OldEnergy = ePointAndChains(Coords[Monomer],Monomer);
   double NewEnergy = ePointAndChains(RNew,Monomer);
#endif
   double de = v[(int)(RNew.zcomp()/WellWidth)] - v[(int)(Coords[Monomer].zcomp()/WellWidth)];

   if (clamped) {
      OldEnergy += VClamp(Coords[Monomer],Monomer);
      NewEnergy += VClamp(RNew,Monomer);
   }

   if (Rand->nextFloat() < exp(Beta*(OldEnergy-NewEnergy-de))) {
      Coords[Monomer] = RNew;
      insert(Monomer);
      return 1;
   }
   else {
      insert(Monomer);
      return 0;
   }
}


int Polymer::jiggle(int num_steps) {
   int NumAccepted=0;

   while (num_steps-- > 0) {
      NumAccepted += jiggle_once();
   }
      
    return NumAccepted;
}

/* Idea of algorithm 
   Forward move, adding to the tail (t), taking off the head (h)
   t        h    ht             ht        
   0123456789 -> 9012345678 -> 8901234567  


   Backward move, i.e. adding to the head (h), taking off the tail (t)
   t        h            ht           ht
   0123456789 -> 1234567890 -> 2345678901 
   */

int Polymer::reptate_once() {
   vect_f  RNew;
   vect_f Rand3d;
   //Subtract off the weight due to the gaussian generator
#ifdef GAUSSIAN
   double Coeff = 0.5*(1.0/(StepLength*StepLength) - 1.0/(MoveSize*MoveSize));
#endif

      int WhichEnd = Rand->nextInt(2);// 0: head is added to, 1: tail is added to
      int WhichMonomer = WhichEnd*(NumMon-1);
      int HeadOrTail = (Head+WhichEnd)%NumMon;// is the end that has a monomer added to it.
      int TailOrHead = (Head+(1-WhichEnd))%NumMon;// The other side, which is chopped
      int AdjacentMonomer = (TailOrHead - (2*WhichEnd-1) + NumMon)%NumMon;//Monomer adjacent TailOrHead
      double *v = polymers->VConfining;
      double WellWidth = polymers->WellWidth;
      double OldEnergy, NewEnergy, de;

      RNew = Coords[HeadOrTail];
      remove(TailOrHead);
#ifdef GAUSSIAN
      Rand3d.set(MoveSize*Rand->nextGaussian(),MoveSize*Rand->nextGaussian(),MoveSize*Rand->nextGaussian());
#else//fixed length
      Rand->nextSphere(Rand3d);
      Rand3d *= StepLength;
#endif

      RNew += Rand3d;

      if (!grid->inBox(RNew))
	 goto REJECT;
#ifdef GAUSSIAN
      OldEnergy = ePointAndChains(Coords[TailOrHead],TailOrHead)+Coeff*Coords[AdjacentMonomer].distsq(Coords[TailOrHead]);
      NewEnergy = ePointAndChains(RNew,TailOrHead)+Coeff*RNew.distsq(Coords[HeadOrTail]);
#else//fixed length
      OldEnergy = ePointAndChains(Coords[TailOrHead],TailOrHead);
      NewEnergy = ePointAndChains(RNew,TailOrHead);
#endif
      de = v[(int)(RNew.zcomp()/WellWidth)] - v[(int)(Coords[TailOrHead].zcomp()/WellWidth)];
      if (Rand->nextFloat() < exp(-Beta*(NewEnergy-OldEnergy+de))) {
	 Head = (Head - 2*WhichEnd + 1 + NumMon)%NumMon;
	 Coords[(Head + WhichEnd)%NumMon] = RNew;
	 insert((Head + WhichEnd)%NumMon);
#ifdef AMOUNT_LEFT
	 mark[Head] = 1;
	 int SumMark=0;
	 for(int k=0; k < NumMon; k++) SumMark += mark[k];
	    if (SumMark == NumMon-1) {
	       printf("inital tube is gone\n");
	    }

#endif
	 return 1;
      }
REJECT:
      insert(TailOrHead);

      return 0;
}


   //reshuffle to correct order before returning
int Polymer::reshuffle() {
   vect_f * TmpCoords = array_copy(Coords,NumMon);

   int Tail = (Head + 1 + NumMon)%NumMon;

   for(int i=0; i < NumMon; i++) {
      remove(i);
   }

   for(int i=0; i < Tail; i++) {
      Coords[i+NumMon-Tail] = TmpCoords[i];
   }

   for(int i=0; i < NumMon-Tail; i++) {
      Coords[i] = TmpCoords[i+Tail];
   }

   for(int i=0; i < NumMon; i++) {
      insert(i);
   }

   delete[] TmpCoords;
   Head = NumMon -1;
   return Head;
}
   

int Polymer::reptate(int num_steps) {
   vect_f  RNew;
   vect_f Rand3d;
   //Subtract off the weight due to the gaussian generator

   double Coeff = 0.5*(1.0/(StepLength*StepLength) - 1.0/(MoveSize*MoveSize));
   int NumAccepted = 0;

   while (num_steps-- > 0) {
      NumAccepted += reptate_once();
   }
   reshuffle();
   return NumAccepted;
}


//For interface ParameterFindable
void Polymer::initQuantities(int numParams) {
   QuantitiesToAverage = new double[numParams];
}

double * Polymer::getAveQuantities() {
   QuantitiesToAverage[0] = (double) ESpring();
   QuantitiesToAverage[1] =  (double) ERepulsive();
   //QuantitiesToAverage[2] =  (double) RgSq();
   return QuantitiesToAverage;
}

//tunes to optimize step size
double Polymer::move(int NumSteps) {
   AcceptRate = 0;
   int AveragingLength = 100;
   int NumIterations = NumSteps/AveragingLength;
   int TruncSteps = AveragingLength*NumIterations;
   int LeftOver = NumSteps - TruncSteps;
   double MinMove = 1.0*pow((double) Beta*Repulsion,(double) 0.25);

   while (NumIterations-- > 0)
   {
      //AcceptRate = ((double) fixedMove(AveragingLength))/AveragingLength;// //real code
      AcceptRate = ((double) fixedMove(NumSteps))/NumSteps;//used for testing
      if (MoveSize > MinMove) 
	 MoveSize = (double) (MoveSize*(0.5 + AcceptRate));
   }
   fixedMove(LeftOver);

   return MoveSize;
}

double Polymer::getAcceptanceRate() {
   return AcceptRate;
}


vect_f * Polymer::getForce(double t, vect_f * noise, vect_f &f_end) {
   //cout << "time in getForce" << t << endl;
   vect_f * force = new vect_f[NumMon];//maybe should keep in Polymer class
   for(int i=0; i < NumMon; i++) {
      force[i] = noise[i];
#ifndef PHANTOM_CHAIN
      force[i] += fPointAndChains(Coords[i],i);
//#else//PHANTOM_CHAIN
//      force[i].set(0.0,0.0,0.0);
#endif//end PHANTOM_CHAIN


#ifdef SPRING_SIMULATION
   force[i] += nnforces(i);
   //end SPRING_SIMULATION
#elif defined(STIFF_SIMULATION)
//   force[i] += nnforces(i);
   force[i] += StiffForces(i);
#endif// STIFF_SIMULATION
   }
#ifdef MUSCLE_SIMULATION
   vect_f EndForces[2];
   vect_f f_rep = beadForce(Coords[0],Coords[NumMon-1]);
   if (Bind1->isBound()) {
      actinForce(Coords[0], Coords[1], EndForces, true, t);
   }
   else {
      EndForces[0] = EndForces[1] = 0;
   }
   force[0] += EndForces[0] + f_rep;
   force[1] += EndForces[1];

   if (Bind2->isBound()) {
      actinForce(Coords[NumMon-1], Coords[NumMon-2], EndForces, false, t);
   }
   else {
      EndForces[0] = EndForces[1] = 0;
   }
   force[NumMon-1] += EndForces[0]-f_rep;
   force[NumMon-2] += EndForces[1];
#endif


   if (ExtForce != 0) {
      vect_f EndForce = extForces();
      force[0] += EndForce;
      force[NumMon-1] -= EndForce;
   }
#ifdef SPRING_SIMULATION
   return force;
#else
   //force[0] = zero_vect;
   f_end = force[0];

   vect_f * f_total = tension->calcForce(Coords, force, Velocities);
   f_end += tension->getTension()[0]*(Coords[1]-Coords[0]);
   delete[] force;
   return f_total;
#endif
}
