#include "RandNumGenerator.h"

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

#if 0
RandNumGenerator::RandNumGenerator() {
   FILE * random_file = fopen("/dev/urandom","r");
   fread(&seed,sizeof(seed),1,random_file);
   fclose(random_file);
   RandNumGenerator(seed);
}
#endif

RandNumGenerator::RandNumGenerator(long seed) {
   iy=0;
   iset=0;
   this->seed = seed;

   for (int j=NTAB+7;j>=0;j--) {
      int k=seed/IQ;
      seed=IA*(seed-k*IQ)-IR*k;
      if (seed < 0) seed += IM;
      if (j < NTAB) iv[j] = seed;
   }
   iy=iv[0];
}

double RandNumGenerator::nextFloat() {
   double temp;
   int k=(seed)/IQ;
   seed=IA*(seed-k*IQ)-IR*k;
   if (seed < 0) seed += IM;
   int j=iy/NDIV;
   iy=iv[j];
   iv[j] = seed;
   if ((temp=AM*iy) > RNMX) return RNMX;
   else return temp;
}

int RandNumGenerator::nextInt(int NumMon) {
   return ((int) (NumMon*nextFloat()));
}

double RandNumGenerator::nextGaussian() {
   double fac,rsq,v1,v2;

   if  (iset == 0) {
      do {
	 v1=2.0*nextFloat()-1.0;
	 v2=2.0*nextFloat()-1.0;
	 rsq=v1*v1+v2*v2;
      } while (rsq >= 1.0 || rsq == 0.0);
      fac=sqrt(-2.0*log(rsq)/rsq);
      gset=v1*fac;
      iset=1;
      return v2*fac;
   } else {
      iset=0;
      return gset;
   }
}

void RandNumGenerator::nextCircle(double &x, double &y) {
   double norm;
   do {
      x= 2.0*nextFloat()-1.0;
      y= 2.0*nextFloat()-1.0;
      norm = x*x+y*y;
   } while (norm > 1.0 || norm == 0);
   norm = 1.0/sqrt(norm);
   x *= norm;
   y *= norm;
}



void RandNumGenerator::nextSphere(vect_f & rand) {
   double x,y,z,norm;
   do {
      x= 2.0*nextFloat()-1.0;
      y= 2.0*nextFloat()-1.0;
      z= 2.0*nextFloat()-1.0;
      norm = x*x+y*y+z*z;
   } while (norm > 1.0 || norm == 0);
   norm = 1.0/sqrt(norm);
   rand.set(norm*x,norm*y,norm*z);

#if 0//quick and dirty:
   rand.set((double) nextGaussian(),(double)  nextGaussian(),(double) nextGaussian()); 
   rand /= sqrt(rand * rand);
#endif
}

vect_f * RandNumGenerator::getGaussVect_f(double StdDev) {
   return (new vect_f((double) (StdDev*this->nextGaussian()),(double) (StdDev*this->nextGaussian()), (double) (StdDev*this->nextGaussian())));
}

void  RandNumGenerator::getGaussVect_f(double StdDev, vect_f &rand) {
   rand.set((double) (StdDev*this->nextGaussian()),(double) (StdDev*this->nextGaussian()), (double) (StdDev*this->nextGaussian()));
}


vect_f * RandNumGenerator::getBoxVect_f(double x, double y, double z) {
   return new vect_f((double) (x*(2*this->nextFloat()-1)),(double) (y*(2*this->nextFloat()-1)), (double) (z*(2*this->nextFloat()-1)));
}


#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

