#include <iostream>
#include <math.h>
#include "Tension.h"

//n = NumMon-1
Tension::Tension(int n, double StepLength, double Damping) {
   delta = new vect_f[n];
   T = new double[n];
   a = new double[n];
   b = new double[n];
   c = new double[n];
   this->Damping = Damping;
   rhs = new double[n+1];//used for the rhs of the matrix equation
   vrhs = new vect_f[n+1];//used for the rhs of eqn of motion ma = f
   this->n = n;
   StepLengthSq = SQR(StepLength);
}

   double * Tension::calcTension(vect_f * r, vect_f * f, vect_f * v) {
   int i;
   for(i=0; i < n; i++) {
      delta[i] = r[i+1]-r[i];
      b[i] = -2.0*delta[i]*delta[i];
   }
   for(i=1; i < n; i++) {
      a[i] = delta[i]*delta[i-1];
   }
    for(i=0; i < (n-1); i++) {
      c[i] = delta[i]*delta[i+1];
   }

   for(i=0; i < n; i++) {
      rhs[i] = (f[i]-f[i+1])*delta[i] - v[i+1].distsq(v[i]) - 20.0*delta[i]*(v[i+1]-v[i]) - (delta[i]*delta[i]-StepLengthSq);
#if 1//Turn off if Damping=0 to make it faster
      rhs[i] += Damping*delta[i]*(v[i+1]-v[i]); 
#endif
   }
   tridag();
   return T;
}

vect_f * Tension::calcForce(vect_f * r, vect_f * f, vect_f * v) {
   calcTension(r,f,v);
   //double max_T = 0;
   for(int i=1; i < n; i++) {
      vrhs[i] = T[i]*delta[i]-T[i-1]*delta[i-1]+f[i];
    //  max_T = max(fabs(T[i]),max_T);
   }
   vrhs[0] = T[0]*delta[0] + f[0];
   vrhs[n] = -T[n-1]*delta[n-1] + f[n];
   //assert (max_T < 50.0);
   return vrhs;
}

/*
 * modified numerical recipes
 * solve Mu=r
 * b main diag, a lower band, b upper band, 
 * b[0]...b[n-1], c[2]...c[n-1], 
 * b[0] c[0] 0     0 ...
 * a[1] b[1] c[1]
 * 0    a[2] b[2] c[2]   0 ...
 * .
 * .
 * .
 * .
 * 0    0 .                     a[n-2] b[n-2] c[n-2]
 * 0    0 .                     0      a[n-1] b[n-1]
 */



void Tension::tridag()
{
   int j;
   double bet,*gam;

   gam= new double[n];

   if (b[0] == 0.0) 
   {
      cout << "Error 1 in tridag" << endl;
      exit(4);
   }
   T[0]=rhs[0]/(bet=b[0]);
   for (j=1;j< n;j++) {
      gam[j]=c[j-1]/bet;
      bet=b[j]-a[j]*gam[j];
      if (bet == 0.0)	
      {
	 cout << "Error 2 in tridag" << endl;
	 exit(4);
      }
      T[j]=(rhs[j]-a[j]*T[j-1])/bet;
   }
   for (j=(n-2);j>=0;j--)
      T[j] -= gam[j+1]*T[j+1];
   delete [] gam;
}

