
from scipy import *
from pylab import *
import numpy as npy
import time

def randsphere(radius = 1.0):
    while 1:
        r = 2.*(random(3) - array([0.5,0.5,0.5]))
        rsq = sum(r*r)
        if rsq < 1:
            return radius*r/sqrt(rsq)

def randcircle(radius = 1.0):
    while 1:
        r = 2.*(npy.random.random(2) - array([0.5,0.5]))
        rsq = sum(r*r)
        if rsq < 1:
            return radius*r/sqrt(rsq)

def rand2sphere(radius = 1.0):
   r = randcircle()
   return array([r[0],r[1],0.0])


def concentration_linear(r):
    return gradient*r[0]

#Concentration is being defined with a negative sign so that it agrees qualitatively with
#the notion of a potential. The smaller the the concentration (potential) the more attractive
#it is to the bacteria:

def concentration_well(r):
    return k*sum(r*r) #A quadratic concentration (potential) well.


#This defines the probabality per unit time of the bacteria tumbling and therefore changing direction:
#Remember that c is being used with the negative of what you'd expect so that it's more like a potential.
#It wants to minimize it's potential.
# But here we'll define prob_tumb with the  delta_c that you'd expect: higher delta_x means less likelihood of tumbling:

def prob_tumble(delta_c):
    #if the change in the concentration delta_c is sufficiently small or negative we'd better tumble with probality 1:
    if delta_c < .5-c_0:
    	return 1
    #For increasing delta_c, make the probability of tumbling per unit time go down:
    return 0.5/(c_0 + delta_c)

def compare_to_histogram(a, Graph=True):
    n, bins, patches = hist(a, 100, normed=1)
    sigma= std(a)
    gauss = exp(-bins*bins/(2*sigma**2))/(sqrt(2*pi)*sigma)
    print "sigma = ", sigma
    if Graph:
        plot(bins,gauss)


y_offset = -3
r_offset = array((0,y_offset,0))

k = 0.01
gradient = 0.01
delta_c = 0
c_0 = 0.5
c_old = 0.
c_new = 0.
r = zeros((3))
#velocity of the bacteria:
v = 1

#change this to 0 to actually gather data:
graphics = 0
if graphics:
    from visual import *


if (graphics):
   cilia_radius = 1.0
   floor = box (pos=(0,y_offset,0), length=400, height=1.0, width=40, color=color.blue)
   rod = cylinder(pos=(0,y_offset,0), axis=(0,1.00,0), radius=cilia_radius, color=color.green)
   ball = sphere (pos=(0,y_offset+1.0,0), radius=3.0, color=color.red)

if graphics:
   ion()

z = array([])

for i in range(1000):
   #Below we'll calculate delta_c, the change in concentration. Of course our definition
   # of c differs by a minus sign, so when we invoke prob_tumble, we'll correct for that:
#   time.sleep(0.001)
   if graphics:
      rate(10000)
   if npy.random.random(1)[0] < prob_tumble(-delta_c):
      dir = rand2sphere() #choose a random direction
   c_old = c_new #remember the old value of the concentration
   r += v*dir # change the position of the bacteria
   #the concentration at this new position:
   c_new = concentration_well(r)
   #The change in c is used above to decide how often to tumble.
   delta_c = c_new -c_old
   z = append(z,r[0])
   if i%1000 == 0:
       if graphics:
           rod.axis = dir
           rod.pos = r_offset + r
           ball.pos = r_offset + r + dir
       elif False:
           plot(z)

cla()
clf()
compare_to_histogram(z)
show()
