#See poisson_boltz_1d.py for for comments on what the code does

from scipy import *
from pylab import *
import pylab
from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.ticker as ticker


C1 = 1/6.0
C2 = 1/12.0

#
DT = 0.2
#DT = 0.5

DT3 = DT/3.0

def diffuse(p, c1=C1, c2=C2, dt= DT3):
    #simpler but anisotropic:
    #p[1:-1,1:-1] = c1*(p[0:-2,1:-1] + p[2:,1:-1] + p[1:-1,0:-2] + p [1:-1,2:]) + c2*(p[0:-2,0:-2] + p[2:,2:] + p[0:-2,2:] + p[2:,0:-2])
    #This is better: (more isotropic)
    p[1:-1,1:-1] =  (1-dt)*p[1:-1,1:-1] + dt*(c1*(p[0:-2,1:-1] + p[2:,1:-1] + p[1:-1,0:-2] + p [1:-1,2:]) + c2*(p[0:-2,0:-2] + p[2:,2:] + p[0:-2,2:] + p[2:,0:-2]))
    return



nx = 32
ny = 32


v = zeros((nx,ny))
v_old = zeros((nx,ny))

line_length = 4
length_scale = 0.25
c_l = length_scale**2
Q0 = -1

#this makes the imshow graph have the x-y right scaling:
extent = [0, (nx-1)*length_scale,0, (ny-1)*length_scale]

ion()


for k in range(50000):
    v_old = v + 0
    diffuse(v)
    v -= c_l*DT*sinh(v)
#   Linearized eqn:
#    v -= c_l*DT*v

    v[:,0] = 0
    v[0,:] = 0
    v[-1,:] = 0
    v[:,-1] = 0
#   this makes a nice collection of charges of varying signs. You can play around with this
#   to create other fixed charge configurations:
    v[nx/2,ny/2-2] -= length_scale*Q0*DT
    v[nx/2-line_length:nx/2+line_length,ny/2] -= length_scale*Q0*DT
    v[nx/2,ny/2+2] -= -length_scale*Q0*DT

    if 1 or  k%50 == 0: 
       imshow(v,extent=extent)
       draw()
       pause(0.0001)
       error = sqrt(sum((v_old-v)**2))
       print "error = ", error
       if error < 0.04:
           break

        #plt.show()


ioff()
#This pops up a 3d graph you can rotate around:
X = np.arange(0,nx)
Y = np.arange(0,ny)
X, Y = np.meshgrid(X, Y)

fig = plt.figure()
ax = Axes3D(fig)
ax.set_axis_off()


ax.plot_surface(X, Y, v, rstride=1, cstride=1, cmap=cm.jet)
plt.show()

