
from scipy import *
from pylab import *
import time

C1 = 1/6.0
C2 = 1/12.0
Frac = 0

def diffuse(p, c1=C1, c2=C2, frac=Frac):
    p[1:-1,1:-1] =  (1-frac)*p[1:-1,1:-1] + frac*(c1*(p[0:-2,1:-1] + p[2:,1:-1] + p[1:-1,0:-2] + p [1:-1,2:]) + c2*(p[0:-2,0:-2] + p[2:,2:] + p[0:-2,2:] + p[2:,0:-2]))
    return

def diffuse_periodic(p, c1=C1, c2=C2, frac=Frac):
    tmp[:,:] = p[:,:]
    p[1:-1,1:-1] =  (1-frac)*p[1:-1,1:-1] + frac*(c1*(p[0:-2,1:-1] + p[2:,1:-1] + p[1:-1,0:-2] + p [1:-1,2:]) + c2*(p[0:-2,0:-2] + p[2:,2:] + p[0:-2,2:] + p[2:,0:-2]))
    p[0,1:-1] = (1.-frac)*p[0,1:-1] + frac*(c1*(p[1,1:-1] + p[-1,1:-1]  + p[0,0:-2] + p [0,2:]) + c2*( p[1,2:] + p[1,0:-2] + p[-1,2:] + p[-1,0:-2]))
    p[1:-1,0] = (1.-frac)*p[1:-1,0] + frac*(c1*(p[1:-1,1] + p[1:-1,-1]  + p[0:-2,0] + p [2:,0]) + c2*( p[2:,1] + p[0:-2,1] + p[2:,-1] + p[0:-2,-1]))
    p[0,0] =  (1.-frac)*p[0,0] + frac*(c1*(p[1,0] + p[0,1]+p[-1,0]+p[0,-1]) + c2* (p[1,1]+p[1,-1]+p[-1,1]+p[-1,-1]))

    p[-1,1:-1] = (1.-frac)*tmp[-1,1:-1] + frac*(c1*(tmp[0,1:-1] + tmp[-2,1:-1]  + tmp[-1,0:-2] + tmp [-1,2:]) + c2*( tmp[0,2:] + tmp[0,0:-2] + tmp[-2,2:] + tmp[-2,0:-2]))
    p[1:-1,-1] = (1.-frac)*tmp[1:-1,-1] + frac*(c1*(tmp[1:-1,0] + tmp[1:-1,-2]  + tmp[0:-2,-1] + tmp [2:,-1]) + c2*( tmp[2:,0] + tmp[0:-2,0] + tmp[2:,-2] + tmp[0:-2,-2]))
    p[0,-1] =  (1.-frac)*tmp[0,-1] + frac*(c1*(tmp[1,-1] + tmp[0,0]+tmp[-1,-1]+tmp[0,-2]) + c2* (tmp[1,0]+tmp[1,-2]+tmp[-1,0]+tmp[-1,-2]))
    p[-1,0] =  (1.-frac)*tmp[-1,0] + frac*(c1*(tmp[-1,1] + tmp[0,0]+tmp[-1,-1]+tmp[-2,0]) + c2* (tmp[0,1]+tmp[-2,1]+tmp[0,-1]+tmp[-2,-1]))
    p[-1,-1] =  (1.-frac)*tmp[-1,-1] + frac*(c1*(tmp[-1,0] + tmp[0,-1]+tmp[-1,-2]+tmp[-2,-1]) + c2* (tmp[0,0]+tmp[-2,0]+tmp[0,-2]+tmp[-2,-2]))

    return



def make_gauss_filter(sigsq,nx,ny): 
   inv = 1/(2.0*sigsq)
   filtery = matrix(array([(exp(-(x**2)*inv) + exp(-((ny-x)**2)*inv)) for x in range(ny)],dtype=float))
   normy = sum(filtery)
   filtery = filtery/normy

   filterx = matrix(array([(exp(-(x**2)*inv) + exp(-((nx-x)**2)*inv)) for x in range(nx)],dtype=float))
   normx = sum(filterx)
   filterx = filterx/normx
   filterx = transpose(filterx)
   return fft2(array(filterx*filtery))
   

nx = 128
ny = 128
dt = 0.002
a=150
b = 100
Du = 10.0
Dv = 500.0
c=1.5
gamma = 2.0

seed(1000)


filter_u = make_gauss_filter(2*Du*dt,nx,ny)
filter_v = make_gauss_filter(2*Dv*dt,nx,ny)

dims = (nx,ny)

u = 38.*ones(dims,dtype=float)
v = 25.*ones(dims,dtype=float)
tmp = zeros(dims,dtype=float)
u += standard_normal(dims)
v += standard_normal(dims)

ion()
for k in range(40000):
   diffuse_periodic(u,frac=dt*Du);
   diffuse_periodic(v,frac=dt*Dv);

   h = 13.0*u*v/(1+u*(1+0.05*u))

   u +=  dt*((a-u) -h)*gamma
   v += dt*(c*(b-v)-h)*gamma
   if (k%100 == 0): 
        imshow(u)
        draw()
        pause(0.0001)

show()


