
from scipy import *
from pylab import *
import time
import sys
from scipy.optimize import leastsq




#technical (can ignore if not a coder)
#get rid of the end of the data if its value is too low because it's quite noisy
def make_cutoff(s, r_min = 0.01):
    out = []
    s0 = s[0]
    y_cut  = s[0]*r_min
    for y in s:
        out.append(y)
        if y < y_cut:
            return out #getting too small so leave keeping only this much data

    return out
     

def generate_white_noise(N,amplitude=1):
   return amplitude*standard_normal(N)


def compare_to_histogram(a, Graph=True):
    n, bins, patches = hist(a, 100, normed=1)
    sigma= std(a)
    gauss = exp(-bins*bins/(2*sigma**2))/(sqrt(2*pi)*sigma)
    print("sigma = ", sigma)
    if Graph:
        xlabel("x")
        ylabel("P(x)")
        plot(bins,gauss)

def compute_correlation(a, Graph=True):
    f = abs(fft(a))**2
    corr = real(ifft(f))/len(f)
    #now fit:
    t_window = 6
    n_window = int(t_window/dt)
    t_range = arange(0,t_window,dt)
    fitf = lambda p, t: p[0]*exp(-t/p[1])
    errorf = lambda p, omega, data: fitf(p,omega)-data 
    p0 = [corr[0],1.0]
    p1, success = leastsq(errorf,p0[:],args=(t_range,corr[0:n_window]))
    print("fitting correlation function: p1 = ", p1)

    if Graph:
        xlabel("t")
        ylabel("<x(0)x(t)>")
        plot(t_range, corr[0:n_window],t_range, fitf(p1,t_range))
    return corr

#Increasing smooth_time say from 300 to 3000 will improve the frequency resolution but will also be more noisy:
def compute_smoothed_power_spect(corr, smooth_time = 300.):
    var = smooth_time**2
    filter = array([exp(-0.5*k**2/var) + exp(-0.5*(k-len(corr))**2/var) for k in range(len(corr))])
    return(dt*real(fft(filter*corr)))

#This is a bit technical. This will fit the power spectrum to a Lorenzian. 
#It's first argument is the power spectrum s, and it fits this using "leastsq" and the function errorf.
#It finds parameters for the Lorentzian that minimizes the difference between the data in s, and
#Its Lorentzian fit. 
def fit_spect(s, Graph=True):
    small = make_cutoff(s)
    s_limit = size(small)
    t_limit = dt*size(s)
    om = (2*pi/t_limit)*arange(0,s_limit,1.0,dtype=float)

    fitf = lambda p, omega: p[0]/(1+omega**2/p[1]**2) 
    errorf = lambda p, omega, data: fitf(p,omega)-data 
    p0 = [s[0],2.0]
    p1, success = leastsq(errorf,p0[:],args=(om,small))
    print(p1)

    n_tot = t_max/dt
    dw = 2*pi/n_tot
    dw_bin = dw*smooth_bins


    if Graph:
        xlabel(r'$\omega$')
        ylabel(r'$|{\hat x}_\omega|^2$')
        plot(om,small,"b^",om,fitf(p1,om),"r-")

def sqrt_lorenzian(N,w_0):
    w = arange(0,N)
    return w_0*sqrt(1/(w**2 + w_0**2)+1/((w-N)**2+w_0**2))

def sharp_cutoff(N,w_0):
    w = ones((N))
    w[w_0:-w_0] = 0
    return w
    
def inv_sharp_cutoff(N,w_0):
    w = zeros((N))
    w[w_0:-w_0] = 1
    return w
    

Seed = 10000
seed(Seed)

#Data is generated for the time range 0 to t_max:
t_max= 10000.
#in increments of dt:
dt = 0.01
#total number of data points:
N = int(t_max/dt)
#Now lets make some white noise:
white = generate_white_noise(N)

#It's fourier transform:
fw = fft(white)

omega = arange(0,N)

#Now lets produce filtered noise:
frac = 1/100.0
omega_0 = N*frac
#Now we're going to attenuate the high frequency components:
fa = fw * sqrt_lorenzian(N,omega_0)
#Use this line instead to explore the behavior with a sharp high frequency cutoff:
#fa = fw * sharp_cutoff(N,omega_0)
# Use this to filter just a band of frequencies.
#fa = fw * inv_sharp_cutoff(N,omega_0*0.9)*sharp_cutoff(N,omega_0)

#Now go from the frequency to the time domain, to get the way the noise evolves in time:
a = real(ifft(fa))

figure(0)
#Lets find a histogram of the amplitudes:
compare_to_histogram(a)

smooth_bins = 400
figure(1)
#Lets compute its correlation function:
corr = compute_correlation(a, Graph=True)
figure(2)
#Now get it's power spectrum:
s_total = compute_smoothed_power_spect(corr)
#And see how well it fits to a Lorentzian (it should since that's how we made it)
#comment out if trying funky filters that won't fit well to prevent errors:
fit_spect(s_total, Graph=True)

NoiseRange = 1000
figure(3)
#How some of the white noise data looks:
xlabel("t")
ylabel("white noise")

plot(dt*arange(0,NoiseRange),white[0:NoiseRange])

figure(4)
#How some of the filtered noise looks:
xlabel("t")
ylabel("filtered noise")
plot(dt*arange(0,NoiseRange),a[0:NoiseRange])


show()


