
from scipy import *
from scipy.linalg import *
from pylab import *
from scipy.integrate import *

#See elect.py (electrophoresis in a constant field) for more of an explanation

class Tension:
    def __init__(self, T, n, d):
        self.T = T
        self.delta = zeros((n-1,d),dtype=double)
        self.delta_f = zeros((n-1,d),dtype=double)
        self.u = zeros((n-1),dtype=double)
        self.l = zeros((n-1),dtype=double)
        self.rhs = zeros((n-1,d),dtype=double)
        self.diag = zeros((n-1),dtype=double)
        self.force = zeros((n,d),dtype=double)


    def tension(self, r, f):
        self.delta = r[1:,:]-r[:-1,:]
        self.diag = -2*sum(self.delta*self.delta,axis=1)
        self.delta_f = f[1:,:]-f[:-1,:]
        u_vect = self.delta[1:,:]*self.delta[:-1,:]
        self.u[1:] = sum(u_vect, axis = 1)
        self.l[:-1] = self.u[1:]
        self.rhs = -sum(self.delta_f*self.delta,axis=1) - c1*(-0.5*self.diag - StepLengthSq)

        return solve_banded((1, 1), array((self.u,self.diag,self.l)), self.rhs, overwrite_ab=0, overwrite_b=0, debug=0)

    def calcForce(self,r,f):
        t = self.tension(r,f)
        for i in range(d):
            self.force[1:-1,i] = t[1:]*self.delta[1:,i] - t[0:-1]*self.delta[0:-1,i] + f[1:-1,i]
        self.force[0,:] = t[0]*self.delta[0,:] + f[0,:]
        self.force[n-1,:] = -t[n-2]*self.delta[n-2,:] + f[n-1,:]
        return self.force

def latticeForce(r):
    distsq[:] = 0
    for i in range(0,d-1):
        rem_r[:,i] = (remainder(a*10000.0+r[:,i],a)/a - 0.5)
        distsq[:] += rem_r[:,i]**2
    for i in range(0,d-1):
       f_lattice[:,i] = 0.01*signbit(distsq[:]-softcoresq) * rem_r[:,i] * (1.0/(distsq[:]-hardcoresq) - oneoverblah)**4
        
    if False and any(f_lattice != 0):
        print(f_lattice)
    return f_lattice


f_ext = array([7,5.0,0.0])

f_forward = array([7,5.0,0.0])
f_backward = array([-7,-5.0,0.0])


# This changes the direction of the applied field depending on the time. It takes 2 steps forwards and one backwards:
def pulsed_f_ext(time, PulseTime = 0.1):
    PulseCount = int_(time/PulseTime)
    phase = PulseCount%3
    if phase == 2:
        return f_backward
    else:
        return f_forward
    

def dx_dt(r_in,t):
    r = reshape(r_in,(n,d))
    f = noise + latticeForce(r) + f_ext
    #f = noise +  f_ext
    return tens.calcForce(r,f).flatten()

def length_error(r):
    delta = r[1:,:] - r[:-1,:]
    dlt_sq = delta*delta
    return sum(sum(dlt_sq)) - (n-1)

def construct_chain(n):
    r = zeros((1,d))
    r1 = zeros((1,d))
    shift = a*array([-10.0,-10.0,0.0])
    for i in range(n-1):
        dlt = array([0.0,1.0,0.0])*(2*(i%2)-1)

        r1 += reshape(dlt,(1,d))
        
        r = append(r, r1, 0)
    r += shift
    return r

   

seed(3001)
StepLength = 1.0
StepLengthSq = StepLength**2
d = 3
n = 32
dt = 0.05
T = 1.0
c1 = 1 #helps to maintain constant link lengths

a = 2.21
f_lattice = zeros((n,d))
distsq = zeros((n))
rem_r = zeros((n,d-1))

hardcore = 0.6/a
hardcoresq = hardcore**2
softcore = 1.0/a
softcoresq = softcore**2
oneoverblah = 1.0/(softcoresq-hardcoresq)


tens = Tension(T, n, d)

r = construct_chain(n)

t = arange(0,2*dt,dt)

f = zeros((n,d))

graphics= True
if graphics:
    from vpython import *

    def vc(r):
        if d== 3:
           return vector(r[0],r[1],r[2])
        elif d==2:
           return vector(r[0],r[1],0)



    ball = []
    cyl = []
    for i in range(n-1):
        cyl.append(cylinder (pos=vc(r[i]), axis=vc(r[i+1]-r[i]), color=color.red, radius = 0.2))

    for x in range(-10,25):
        for y in range(-10,30):
            cylinder(pos=vector(a*(x+0.5),a*(y+0.5),0), axis =vector(0,0,5.0), color=color.green, radius=hardcore)
    pointer = arrow(pos=vector(-10,-10,0), axis=vc(f_ext))

if graphics:
    ion()


x_out = r.flatten()

x_out = reshape(x_out,(1,n*d))

time = 0;
for t_coarse in arange(0,100,dt):
    if graphics:
         rate(10000)
#   This is the difference between this file and the constant field case elect.py:
    f_ext = pulsed_f_ext(t_coarse, PulseTime= 1.5)
    t = array([t_coarse,t_coarse+dt])
    noise = sqrt(2*T/dt)*standard_normal((n,d))
    x_out = odeint(dx_dt, x_out[-1] , t, full_output=False)
    time += 1
    if time%300 == 0 : 
        print(length_error(reshape(x_out[-1],(n,d))))
    if  time%10 == 0 and graphics:
        r_out = reshape(x_out[-1],(n,d))
        pointer.axis = vc(f_ext)
        for i in range(n-1):
            cyl[i].pos = vc(r_out[i])
            cyl[i].axis = vc(r_out[i+1]-r_out[i])



