import numpy
from scipy import *
from pylab import *
import time
import weave
from weave import converters
import ParTemp

#See the comments in lattice_protein.py from the previous problem for an explanation
#of the code.

graphics = True

seed(1004)

class LattProt:
    def __init__(self,Sigma,Beta=0.5,box_size=32):
        self.Beta = Beta
        self.box_size = box_size
        self.Sigma = Sigma
        self.R = array(list(range(0,l)), dtype=int) #1d representation of 3d coords r:
        self.Rmin = array(list(range(0,l)), dtype=int) #1d representation of 3d coords rmin:
        self.E = zeros((4),dtype=float)
        self.r = zeros((l,3), dtype=int)
        self.rmin = zeros((l,3), dtype=int)
        self.b_e = backbone_energy(Sigma)
        init_chain(self.R,self.r,box_size)
        self.E[0] = self.Energy()


    def move(self,NumIter):
        R = self.R
        Sigma = self.Sigma
        Rmin = self.Rmin
        Beta = self.Beta
        E = self.E
       

        code = """
        #line 36 "lattice_protein.py"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define NN 6
#define ran() rand()/(RAND_MAX+1.0)
#define en(i,r_,e_) do {                                           \
    int i_;                                                        \
    e_=0;                                                          \
    for(i_=0;i_ < NN;i_++)                                         \
    {                                                              \
        e_ += potts[sigma[i]][sigma[bead[(r_+delta[i_])&c31]]];    \
    }                                                              \
} while(0)
        int cubesize, c1, c2;
        int c3,c31,lm1, *sigma, *bead, *r,*rmin;  
        int count,i,j,m,n,r1,r2,rold1,rold2,r_new,nstates=2;
        int delta[6];
        int w[3][4];
        float eold, **potts;
        float beta,de,energy, energy_min;
        beta = Beta;
        c1 = box_size;
        c2 = c1*c1; 
        c3=c2*c1; 
        c31=c3-1;
        delta[0] = 1;
        delta[1] = -1;
        delta[2] = c1;
        delta[3] = -c1;
        delta[4] = c2;
        delta[5] = -c2;
        w[0][0]=c1;w[0][1]=-c1;w[0][2]=c2;w[0][3]=-c2;
        w[1][0]=1; w[1][1]=-1; w[1][2]=c2;w[1][3]=-c2;
        w[2][0]=1; w[2][1]=-1; w[2][2]=c1;w[2][3]=-c1;
        potts = (float **)  malloc((nstates+3)*sizeof(potts[0]))+2;
        for(i=-2;i< (nstates+1);i++)
            potts[i] =  (float *) calloc((nstates+3),sizeof(potts[i][0]))+2;
        for(i=1;i<3;i++)
            for(j=1;j<3;j++)
                potts[i][j] = V(i,j);
        r = (int *) calloc(l,sizeof(r[0]));
        rmin = (int *) calloc(l,sizeof(r[0]));
        sigma = (int *) calloc(l+2,sizeof(sigma[0])) + 2;
        bead = (int *) malloc(c3*sizeof(bead[0]));
        memset(bead,-1,c3*sizeof(bead[0]));
        lm1 = l-1;
        energy = E(0);
        energy_min = energy;
        for(i=0;i<l;i++)
        {
            sigma[i] = Sigma(i);
            r[i] = R(i);
            bead[r[i]&c31] = i;
        }
        for(count=0;count < NumIter; count++)
        for(n=0;n < l;n++)
        {
            if (energy <= energy_min)
            {
                energy_min = energy;
                memcpy(rmin,r,l*sizeof(r[0]));
            }
            if (n == 0) /* move tail*/
            {
                i = (int)(6.0*ran());
                r1 = r[1]+delta[i];
                if (bead[r1&c31] == -1)
                {
                    float e;
                    en(0,r[0],eold);
                    bead[r[0]&c31]=-1;
                    bead[r1&c31] = 0;
                    rold1=r[0];
                    r[0]=r1;

                    en(0,r1,e);
                    de = e-eold;
                    if (exp(-beta*de) > ran())
                    {
                        energy += de;
                        continue;
                    }
                    bead[r1&c31] = -1;
                    r[0] = rold1;
                    bead[rold1&c31] = 0;
                }
            }
            else if (n == lm1) /* move head*/
            {
                i = (int)(6.0*ran());
                r1 = r[lm1-1]+delta[i];
                if (bead[r1&c31] == -1)
                {
                    float e;
                    en(lm1,r[lm1],eold);
                    bead[r[lm1]&c31]=-1;
                    bead[r1&c31] = lm1;
                    rold1 = r[lm1];
                    r[lm1]=r1;
                    en(lm1,r1,e);
                    de = e-eold;
                    if (exp(-beta*de) > ran())
                    {
                        energy += de;
                        continue;
                    }
                    bead[r1&c31]=-1;
                    r[lm1] = rold1;
                    bead[rold1&c31]=lm1;
                }
            }
            else 
            {
                if (  (n > 1 && n < l-1 && ran() < 0.5))
                {/* choose top hat move */
                    m=n-2;
                    i=abs(r[m+3]-r[m]);
                    if ((i == 1) || (i == c1) || (i == c2))
                    {
                        int dr;
                        float rnd;
                        j=0;
                        while (1)
                        {
                             i = i/c1;
                             if (i == 0)
                                 break;
                             j += 1;
                        }
                        //for(j=0; (i = i/c1) != 0; j++);
                        rnd = ran();
                        dr = w[j][(int)(4*rnd)];
                        r1=r[m]+dr;
                        E(4) = rnd;
                        r2=r1+r[m+3]-r[m];
                        E(3) = 1;
                        if (bead[r1&c31] == -1 && bead[r2&c31] == -1){
                            E(2) = 1;
                            float e1,e2;
                            en(m+1,r[m+1],e1);
                            en(m+2,r[m+2],e2);
                            eold = e1+e2;
                            bead[r[m+1]&c31]=bead[r[m+2]&c31] = -1;
                            bead[r1&c31] = m+1;
                            bead[r2&c31] = m+2;
                            rold1=r[m+1];
                            rold2=r[m+2];
                            r[m+1] = r1;
                            r[m+2] = r2;
                            en(m+1,r1,e1);
                            en(m+2,r2,e2);
                            de =  e1 + e2 - eold;
                            if (exp(-beta*de) > ran())
                            {
                                energy += de;
                                continue;
                            }
                            bead[r1&c31]=bead[r2&c31]=-1;
                            r[m+1] = rold1;
                            r[m+2] = rold2;
                            bead[rold1&c31] = m+1;
                            bead[rold2&c31] = m+2;
                        }
                    }

                }
                else  /*choose right angle move */
                {
                    m=n-1;
                    r_new = r[m+2]+r[m] - r[m+1];
                    if (bead[r_new&c31] == -1)
                    {
                        float e;
                        en(m+1,r[m+1],eold);
                        bead[r[m+1]&c31]=-1;
                        bead[r_new&c31]= m+1 ;
                        rold1=r[m+1];
                        r[m+1] = r_new;
                        en(m+1,r_new,e);
                        de = e-eold;
                        if (exp(-beta*de) > ran())
                        {
                            energy += de;
                            continue;
                        }
                        bead[r_new&c31]=-1;
                        r[m+1]=rold1;
                        bead[rold1&c31]=m+1;
                    }
                }

            }
        }



        for(i=-2;i< (nstates+1);i++)
            free(potts[i]-2);
        free(potts-2);

        free(bead);
        free(sigma-2);
        for(i=0;i<l;i++)
        {
           R(i) = r[i];
           Rmin(i) = rmin[i];
        }
        free(r);
        free(rmin);
        E(0) = energy;
        E(1) = energy_min;
     

        """

        weave.inline(code, ['Sigma', 'R', 'Rmin', 'l','box_size','Beta','V','NumIter','E'], 
        type_converters=converters.blitz, compiler = 'gcc')
        return

#calculates the minimum energy attained so far for this system
    def Energy_min(self):
        return self.E[1]
        
#should be the same as Energ(self)
    def Energy1(self):
        return self.E[0]

    def Energy(self):
        self.convert_1d_to_3d(self.R,self.r)
        r = self.r
        Sigma = self.Sigma
        e = 0
        l = size(Sigma)
        for i in range(l):
            for j in range(i+1,l):
                delta = sum((r[i]-r[j])**2) - 1
                #if abs(delta) < 1:
                #    print "i,j,delta = ", i,j ,delta
                if sum((r[i]-r[j])**2) ==1:
                    e += V[Sigma[i]][Sigma[j]]
                #    if V[Sigma[i]][Sigma[j]] < 0:
                #        print "in energy: i = ", i , "j = ", j
        return e - self.b_e

            
    def convert_1d_to_3d(self,R,r):
        box_size = self.box_size
        r[0,:] = 0
        l = size(R)
        for i in range(1,l):
            delta = R[i]-R[i-1]
            if abs(delta) < box_size:
                r[i] = r[i-1] + delta*dir[0]
            elif abs(delta) < box_size**2:
                r[i] = r[i-1] + int(delta//box_size)*dir[2]
            else:
                r[i] = r[i-1] + int(delta//(box_size*box_size))*dir[4]
        return

def init_chain(R,r,box_size):
    index = 0
    R[0] = 0
    line_length = box_size//2
    direction = 1
    while True:
        for j in range(line_length):
            R[index + 1] = R[index] + direction
            index += 1
            if index == l-1:
                return
        R[index + 1] = R[index] + box_size
        index += 1
        if index == l-1:
            return
        R[index + 1] = R[index] +  box_size
        index += 1
        if index == l-1:
            return
        direction *= -1
    return

def backbone_energy(Sigma):
    e=0
    l = size(Sigma)
    for i in range(l-1):
        e += V[Sigma[i]][Sigma[i+1]]
    return e



dir = array([ [1,0,0], [-1,0,0], [0,1,0], [0,-1,0],[0,0,1],[0,0,-1]], dtype = int)
#seq 1:
#Sigma = array([1, 2, 1, 1, 2, 2, 1, 1, 1, 1, 2, 1, 1, 1, 2, 2, 1, 1, 2, 2, 1, 2, 1, 1, 1, 2, 1, 2, 1, 1, 2, 2, 1, 1, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1], dtype = int)
#Sigma = array([1,1,1,1,2,1,1,2,1,1,1,1,1,2,2,1,2,2,1,1,2,2,1,2,2,2,2,2,2,1,2,2,1,2,2,2,1,2,2,1,1,2,2,1,1,1,2,1], dtype = int)
#seq 10
Sigma = array([2,1,1,2,2,2,2,2,2,1,1,2,2,2,1,1,1,2,1,2,2,1,2,1,1,2,2,1,2,2,1,2,2,1,1,2,2,1,1,1,1,1,1,1,2,2,1,1], dtype = int)
#l = 16
l = size(Sigma)
print("length l = ", l)
Beta = 3.0
box_size = 32
#Sigma = ones((l),dtype=int)
Rglobalmin = array(list(range(0,l)), dtype=int) #1d representation of 3d coords rmin:
           

V = zeros((3,3))
V[1,1] = -1.0
#V[1,2] = -3.3
#V[2,2] = -2.6

lp = LattProt(Sigma,Beta=0.5,box_size=32)
partemp = ParTemp.ParTemp(System=lp, NumSystems = 16, BetaMin = 0.5, BetaMax=8.0,PrintInterval=1,StepsPerMove=10000,NumMoves=1000)
ion()
partemp.graph_on = 1
partemp.move()


